package me.eater.threedom.dom

import me.eater.threedom.utils.joml.times
import org.joml.Matrix4d
import org.joml.Matrix4dc
import java.util.concurrent.atomic.AtomicLong

interface INode<T : INode<T>> : Comparable<INode<*>>, INodeContainer {
    var id: String?
    val classList: MutableSet<String>
    val nodeId: Long
    val parentNode: INode<*>?
    val document: IDocument?
    val absolute: Matrix4dc
        get() = (parentNode?.absolute ?: Matrix4d()) * model

    var model: Matrix4dc

    fun clone(deep: Boolean): T
    fun updateParent(refNode: INode<*>?): Boolean
    fun detachFromDocument()
    fun hasParent(node: INode<*>): Boolean {
        var current: INode<*>? = parentNode;

        while (current != null) {
            if (node.nodeId == current.nodeId) {
                return true
            }

            current = current.parentNode
        }

        return false
    }

    override fun compareTo(other: INode<*>): Int = this.nodeId.compareTo(other.nodeId)
    fun updateAbsolute()

    companion object {
        private val atomicNodeId = AtomicLong(0)
        fun getNextNodeId() = atomicNodeId.getAndIncrement()
    }
}
